@@ -13,7 +13,7 @@
 //
 use std::{
     borrow::Cow,
-    collections::HashMap,
+    collections::{HashMap, HashSet},
     sync::{atomic::Ordering, Arc},
 };
 
@@ -35,19 +35,108 @@ use crate::{
     net::routing::{
         dispatcher::{
             face::FaceState,
-            interests::RemoteInterest,
             pubsub::SubscriberInfo,
             resource::{NodeId, Resource, SessionContext},
             tables::{Route, RoutingExpr, Tables},
         },
         hat::{
-            p2p_peer::initial_interest, CurrentFutureTrait, HatPubSubTrait, SendDeclare, Sources,
+            p2p_peer::{initial_interest, INITIAL_INTEREST_ID},
+            CurrentFutureTrait, HatPubSubTrait, SendDeclare, Sources,
         },
         router::RouteBuilder,
         RoutingContext,
     },
 };
 
+#[inline]
+fn maybe_register_local_subscriber(
+    dst_face: &mut Arc<FaceState>,
+    res: &Arc<Resource>,
+    initial_interest: Option<InterestId>,
+    send_declare: &mut SendDeclare,
+) {
+    if face_hat!(dst_face).local_subs.contains_simple_resource(res) {
+        return;
+    }
+    let (should_notify, simple_interests) = match initial_interest {
+        Some(interest) => (true, HashSet::from([interest])),
+        None => face_hat!(dst_face)
+            .remote_interests
+            .iter()
+            .filter(|(_, i)| i.options.subscribers() && i.matches(res))
+            .fold(
+                (false, HashSet::new()),
+                |(_, mut simple_interests), (id, i)| {
+                    if !i.options.aggregate() {
+                        simple_interests.insert(*id);
+                    }
+                    (true, simple_interests)
+                },
+            ),
+    };
+
+    if !should_notify {
+        return;
+    }
+    let face_hat_mut = face_hat_mut!(dst_face);
+    let (_, subs_to_notify) = face_hat_mut.local_subs.insert_simple_resource(
+        res.clone(),
+        SubscriberInfo,
+        || face_hat_mut.next_id.fetch_add(1, Ordering::SeqCst),
+        simple_interests,
+    );
+
+    for update in subs_to_notify {
+        let key_expr = Resource::decl_key(
+            &update.resource,
+            dst_face,
+            super::push_declaration_profile(dst_face),
+        );
+        send_declare(
+            &dst_face.primitives,
+            RoutingContext::with_expr(
+                Declare {
+                    interest_id: None,
+                    ext_qos: ext::QoSType::DECLARE,
+                    ext_tstamp: None,
+                    ext_nodeid: ext::NodeIdType::DEFAULT,
+                    body: DeclareBody::DeclareSubscriber(DeclareSubscriber {
+                        id: update.id,
+                        wire_expr: key_expr.clone(),
+                    }),
+                },
+                update.resource.expr().to_string(),
+            ),
+        );
+    }
+}
+
+#[inline]
+fn maybe_unregister_local_subscriber(
+    face: &mut Arc<FaceState>,
+    res: &Arc<Resource>,
+    send_declare: &mut SendDeclare,
+) {
+    for update in face_hat_mut!(face).local_subs.remove_simple_resource(res) {
+        send_declare(
+            &face.primitives,
+            RoutingContext::with_expr(
+                Declare {
+                    interest_id: None,
+                    ext_qos: ext::QoSType::DECLARE,
+                    ext_tstamp: None,
+                    ext_nodeid: ext::NodeIdType::DEFAULT,
+                    body: DeclareBody::UndeclareSubscriber(UndeclareSubscriber {
+                        id: update.id,
+                        ext_wire_expr: WireExprType::null(),
+                    }),
+                },
+                update.resource.expr().to_string(),
+            ),
+        );
+    }
+}
+
 #[inline]
 fn propagate_simple_subscription_to(
     _tables: &mut Tables,
@@ -58,75 +147,13 @@ fn propagate_simple_subscription_to(
     send_declare: &mut SendDeclare,
 ) {
     if (src_face.id != dst_face.id)
-        && !face_hat!(dst_face).local_subs.contains_key(res)
+        && !face_hat!(dst_face).local_subs.contains_simple_resource(res)
         && (src_face.whatami == WhatAmI::Client || dst_face.whatami == WhatAmI::Client)
     {
         if dst_face.whatami != WhatAmI::Client {
-            let id = face_hat!(dst_face).next_id.fetch_add(1, Ordering::SeqCst);
-            face_hat_mut!(dst_face).local_subs.insert(res.clone(), id);
-            let key_expr =
-                Resource::decl_key(res, dst_face, super::push_declaration_profile(dst_face));
-            send_declare(
-                &dst_face.primitives,
-                RoutingContext::with_expr(
-                    Declare {
-                        interest_id: None,
-                        ext_qos: ext::QoSType::DECLARE,
-                        ext_tstamp: None,
-                        ext_nodeid: ext::NodeIdType::DEFAULT,
-                        body: DeclareBody::DeclareSubscriber(DeclareSubscriber {
-                            id,
-                            wire_expr: key_expr,
-                        }),
-                    },
-                    res.expr().to_string(),
-                ),
-            );
+            maybe_register_local_subscriber(dst_face, res, Some(INITIAL_INTEREST_ID), send_declare);
         } else {
-            let matching_interests = face_hat!(dst_face)
-                .remote_interests
-                .values()
-                .filter(|i| i.options.subscribers() && i.matches(res))
-                .cloned()
-                .collect::<Vec<_>>();
-
-            for RemoteInterest {
-                res: int_res,
-                options,
-                ..
-            } in matching_interests
-            {
-                let res = if options.aggregate() {
-                    int_res.as_ref().unwrap_or(res)
-                } else {
-                    res
-                };
-                if !face_hat!(dst_face).local_subs.contains_key(res) {
-                    let id = face_hat!(dst_face).next_id.fetch_add(1, Ordering::SeqCst);
-                    face_hat_mut!(dst_face).local_subs.insert(res.clone(), id);
-                    let key_expr = Resource::decl_key(
-                        res,
-                        dst_face,
-                        super::push_declaration_profile(dst_face),
-                    );
-                    send_declare(
-                        &dst_face.primitives,
-                        RoutingContext::with_expr(
-                            Declare {
-                                interest_id: None,
-                                ext_qos: ext::QoSType::DECLARE,
-                                ext_tstamp: None,
-                                ext_nodeid: ext::NodeIdType::DEFAULT,
-                                body: DeclareBody::DeclareSubscriber(DeclareSubscriber {
-                                    id,
-                                    wire_expr: key_expr,
-                                }),
-                            },
-                            res.expr().to_string(),
-                        ),
-                    );
-                }
-            }
+            maybe_register_local_subscriber(dst_face, res, None, send_declare);
         }
     }
 }
@@ -214,7 +241,7 @@ fn declare_simple_subscription(
                             }),
                         },
                         res.expr().to_string(),
-                    ))
+                    ));
             }
         }
     }
@@ -234,67 +261,13 @@ fn simple_subs(res: &Arc<Resource>) -> Vec<Arc<FaceState>> {
         .collect()
 }
 
-#[inline]
-fn remote_simple_subs(res: &Arc<Resource>, face: &Arc<FaceState>) -> bool {
-    res.session_ctxs
-        .values()
-        .any(|ctx| ctx.face.id != face.id && ctx.subs.is_some())
-}
-
 fn propagate_forget_simple_subscription(
     tables: &mut Tables,
     res: &Arc<Resource>,
     send_declare: &mut SendDeclare,
 ) {
     for mut face in tables.faces.values().cloned() {
-        if let Some(id) = face_hat_mut!(&mut face).local_subs.remove(res) {
-            send_declare(
-                &face.primitives,
-                RoutingContext::with_expr(
-                    Declare {
-                        interest_id: None,
-                        ext_qos: ext::QoSType::DECLARE,
-                        ext_tstamp: None,
-                        ext_nodeid: ext::NodeIdType::DEFAULT,
-                        body: DeclareBody::UndeclareSubscriber(UndeclareSubscriber {
-                            id,
-                            ext_wire_expr: WireExprType::null(),
-                        }),
-                    },
-                    res.expr().to_string(),
-                ),
-            );
-        }
-        for res in face_hat!(face)
-            .local_subs
-            .keys()
-            .cloned()
-            .collect::<Vec<Arc<Resource>>>()
-        {
-            if !res.context().matches.iter().any(|m| {
-                m.upgrade()
-                    .is_some_and(|m| m.context.is_some() && remote_simple_subs(&m, &face))
-            }) {
-                if let Some(id) = face_hat_mut!(&mut face).local_subs.remove(&res) {
-                    send_declare(
-                        &face.primitives,
-                        RoutingContext::with_expr(
-                            Declare {
-                                interest_id: None,
-                                ext_qos: ext::QoSType::DECLARE,
-                                ext_tstamp: None,
-                                ext_nodeid: ext::NodeIdType::DEFAULT,
-                                body: DeclareBody::UndeclareSubscriber(UndeclareSubscriber {
-                                    id,
-                                    ext_wire_expr: WireExprType::null(),
-                                }),
-                            },
-                            res.expr().to_string(),
-                        ),
-                    );
-                }
-            }
-        }
+        maybe_unregister_local_subscriber(&mut face, res, send_declare);
     }
 }
 
@@ -315,55 +288,8 @@ pub(super) fn undeclare_simple_subscription(
         }
 
         if simple_subs.len() == 1 {
-            let mut face = &mut simple_subs[0];
-            if let Some(id) = face_hat_mut!(face).local_subs.remove(res) {
-                send_declare(
-                    &face.primitives,
-                    RoutingContext::with_expr(
-                        Declare {
-                            interest_id: None,
-                            ext_qos: ext::QoSType::DECLARE,
-                            ext_tstamp: None,
-                            ext_nodeid: ext::NodeIdType::DEFAULT,
-                            body: DeclareBody::UndeclareSubscriber(UndeclareSubscriber {
-                                id,
-                                ext_wire_expr: WireExprType::null(),
-                            }),
-                        },
-                        res.expr().to_string(),
-                    ),
-                );
-            }
-            for res in face_hat!(face)
-                .local_subs
-                .keys()
-                .cloned()
-                .collect::<Vec<Arc<Resource>>>()
-            {
-                if !res.context().matches.iter().any(|m| {
-                    m.upgrade()
-                        .is_some_and(|m| m.context.is_some() && remote_simple_subs(&m, face))
-                }) {
-                    if let Some(id) = face_hat_mut!(&mut face).local_subs.remove(&res) {
-                        send_declare(
-                            &face.primitives,
-                            RoutingContext::with_expr(
-                                Declare {
-                                    interest_id: None,
-                                    ext_qos: ext::QoSType::DECLARE,
-                                    ext_tstamp: None,
-                                    ext_nodeid: ext::NodeIdType::DEFAULT,
-                                    body: DeclareBody::UndeclareSubscriber(UndeclareSubscriber {
-                                        id,
-                                        ext_wire_expr: WireExprType::null(),
-                                    }),
-                                },
-                                res.expr().to_string(),
-                            ),
-                        );
-                    }
-                }
-            }
+            let face = &mut simple_subs[0];
+            maybe_unregister_local_subscriber(face, res, send_declare);
         }
     }
 }
@@ -409,129 +335,114 @@ pub(super) fn pubsub_new_face(
     }
 }
 
-#[inline]
-fn make_sub_id(res: &Arc<Resource>, face: &mut Arc<FaceState>, mode: InterestMode) -> u32 {
-    if mode.future() {
-        if let Some(id) = face_hat!(face).local_subs.get(res) {
-            *id
-        } else {
-            let id = face_hat!(face).next_id.fetch_add(1, Ordering::SeqCst);
-            face_hat_mut!(face).local_subs.insert(res.clone(), id);
-            id
-        }
-    } else {
-        0
-    }
+fn get_subscribers_matching_resource<'a>(
+    tables: &'a Tables,
+    face: &Arc<FaceState>,
+    res: Option<&'a Arc<Resource>>,
+) -> impl Iterator<Item = &'a Arc<Resource>> {
+    let face_id = face.id;
+
+    tables
+        .faces
+        .values()
+        .filter(move |f| f.id != face_id)
+        .flat_map(|f| face_hat!(f).remote_subs.values())
+        .filter(move |sub| sub.context.is_some() && res.map(|r| sub.matches(r)).unwrap_or(true))
 }
 
 pub(super) fn declare_sub_interest(
     tables: &mut Tables,
     face: &mut Arc<FaceState>,
-    id: InterestId,
+    interest_id: InterestId,
     res: Option<&mut Arc<Resource>>,
     mode: InterestMode,
     aggregate: bool,
     send_declare: &mut SendDeclare,
 ) {
-    if mode.current() && face.whatami == WhatAmI::Client {
-        let interest_id = Some(id);
-        if let Some(res) = res.as_ref() {
-            if aggregate {
-                if tables.faces.values().any(|src_face| {
-                    src_face.id != face.id
-                        && face_hat!(src_face)
-                            .remote_subs
-                            .values()
-                            .any(|sub| sub.context.is_some() && sub.matches(res))
-                }) {
-                    let id = make_sub_id(res, face, mode);
-                    let wire_expr =
-                        Resource::decl_key(res, face, super::push_declaration_profile(face));
-                    send_declare(
-                        &face.primitives,
-                        RoutingContext::with_expr(
-                            Declare {
-                                interest_id,
-                                ext_qos: ext::QoSType::DECLARE,
-                                ext_tstamp: None,
-                                ext_nodeid: ext::NodeIdType::DEFAULT,
-                                body: DeclareBody::DeclareSubscriber(DeclareSubscriber {
-                                    id,
-                                    wire_expr,
-                                }),
-                            },
-                            res.expr().to_string(),
-                        ),
+    if face.whatami != WhatAmI::Client {
+        return;
+    }
+
+    let res = res.map(|r| r.clone());
+    let mut matching_subs = get_subscribers_matching_resource(tables, face, res.as_ref());
+
+    if aggregate && (mode.current() || mode.future()) {
+        if let Some(aggregated_res) = &res {
+            let (resource_id, sub_info) = if mode.future() {
+                let face_hat_mut = face_hat_mut!(face);
+                for sub in matching_subs {
+                    face_hat_mut.local_subs.insert_simple_resource(
+                        sub.clone(),
+                        SubscriberInfo,
+                        || face_hat_mut.next_id.fetch_add(1, Ordering::SeqCst),
+                        HashSet::new(),
                     );
                 }
+                let face_hat_mut = face_hat_mut!(face);
+                face_hat_mut.local_subs.insert_aggregated_resource(
+                    aggregated_res.clone(),
+                    || face_hat_mut.next_id.fetch_add(1, Ordering::SeqCst),
+                    HashSet::from_iter([interest_id]),
+                )
             } else {
-                for src_face in tables
-                    .faces
-                    .values()
-                    .filter(|f| f.id != face.id)
-                    .cloned()
-                    .collect::<Vec<Arc<FaceState>>>()
-                {
-                    for sub in face_hat!(src_face).remote_subs.values() {
-                        if sub.context.is_some() && sub.matches(res) {
-                            let id = make_sub_id(sub, face, mode);
-                            let wire_expr = Resource::decl_key(
-                                sub,
-                                face,
-                                super::push_declaration_profile(face),
-                            );
-                            send_declare(
-                                &face.primitives,
-                                RoutingContext::with_expr(
-                                    Declare {
-                                        interest_id,
-                                        ext_qos: ext::QoSType::DECLARE,
-                                        ext_tstamp: None,
-                                        ext_nodeid: ext::NodeIdType::DEFAULT,
-                                        body: DeclareBody::DeclareSubscriber(DeclareSubscriber {
-                                            id,
-                                            wire_expr,
-                                        }),
-                                    },
-                                    sub.expr().to_string(),
-                                ),
-                            );
-                        }
-                    }
-                }
-            }
-        } else {
-            for src_face in tables
-                .faces
-                .values()
-                .cloned()
-                .collect::<Vec<Arc<FaceState>>>()
-            {
-                if src_face.id != face.id {
-                    for sub in face_hat!(src_face).remote_subs.values() {
-                        let id = make_sub_id(sub, face, mode);
-                        let wire_expr =
-                            Resource::decl_key(sub, face, super::push_declaration_profile(face));
-                        send_declare(
-                            &face.primitives,
-                            RoutingContext::with_expr(
-                                Declare {
-                                    interest_id,
-                                    ext_qos: ext::QoSType::DECLARE,
-                                    ext_tstamp: None,
-                                    ext_nodeid: ext::NodeIdType::DEFAULT,
-                                    body: DeclareBody::DeclareSubscriber(DeclareSubscriber {
-                                        id,
-                                        wire_expr,
-                                    }),
-                                },
-                                sub.expr().to_string(),
-                            ),
-                        );
-                    }
-                }
+                (0, matching_subs.next().map(|_| SubscriberInfo))
+            };
+            if mode.current() && sub_info.is_some() {
+                // send declare only if there is at least one resource matching the aggregate
+                let wire_expr =
+                    Resource::decl_key(aggregated_res, face, super::push_declaration_profile(face));
+                send_declare(
+                    &face.primitives,
+                    RoutingContext::with_expr(
+                        Declare {
+                            interest_id: Some(interest_id),
+                            ext_qos: ext::QoSType::DECLARE,
+                            ext_tstamp: None,
+                            ext_nodeid: ext::NodeIdType::DEFAULT,
+                            body: DeclareBody::DeclareSubscriber(DeclareSubscriber {
+                                id: resource_id,
+                                wire_expr,
+                            }),
+                        },
+                        aggregated_res.expr().to_string(),
+                    ),
+                );
             }
         }
+    } else if !aggregate && mode.current() {
+        for sub in matching_subs {
+            let resource_id = if mode.future() {
+                let face_hat_mut = face_hat_mut!(face);
+                face_hat_mut
+                    .local_subs
+                    .insert_simple_resource(
+                        sub.clone(),
+                        SubscriberInfo,
+                        || face_hat_mut.next_id.fetch_add(1, Ordering::SeqCst),
+                        HashSet::from([interest_id]),
+                    )
+                    .0
+            } else {
+                0
+            };
+            let wire_expr = Resource::decl_key(sub, face, super::push_declaration_profile(face));
+            send_declare(
+                &face.primitives,
+                RoutingContext::with_expr(
+                    Declare {
+                        interest_id: Some(interest_id),
+                        ext_qos: ext::QoSType::DECLARE,
+                        ext_tstamp: None,
+                        ext_nodeid: ext::NodeIdType::DEFAULT,
+                        body: DeclareBody::DeclareSubscriber(DeclareSubscriber {
+                            id: resource_id,
+                            wire_expr,
+                        }),
+                    },
+                    sub.expr().to_string(),
+                ),
+            );
+        }
     }
 }
 
@@ -694,9 +605,7 @@ impl HatPubSubTrait for HatCode {
         key_expr: &KeyExpr<'_>,
     ) -> HashMap<usize, Arc<FaceState>> {
         let mut matching_subscriptions = HashMap::new();
-        if key_expr.ends_with('/') {
-            return matching_subscriptions;
-        }
+
         tracing::trace!("get_matching_subscriptions({})", key_expr,);
         let res = Resource::get_resource(&tables.root_res, key_expr);
         let matches = res
