@@ -19,7 +19,7 @@ use std::{
 
 use petgraph::graph::NodeIndex;
 use zenoh_protocol::{
-    core::{key_expr::OwnedKeyExpr, WhatAmI, ZenohIdProto},
+    core::{WhatAmI, ZenohIdProto},
     network::{
         declare::{
             common::ext::WireExprType, ext, Declare, DeclareBody, DeclareSubscriber, SubscriberId,
@@ -41,12 +41,14 @@ use crate::{
         routing::{
             dispatcher::{
                 face::FaceState,
-                interests::RemoteInterest,
                 pubsub::SubscriberInfo,
                 resource::{NodeId, Resource, SessionContext},
                 tables::{Route, RoutingExpr, Tables},
             },
-            hat::{CurrentFutureTrait, HatPubSubTrait, SendDeclare, Sources},
+            hat::{
+                router::INITIAL_INTEREST_ID, CurrentFutureTrait, HatPubSubTrait, SendDeclare,
+                Sources,
+            },
             router::{disable_matches_data_routes, RouteBuilder},
             RoutingContext,
         },
@@ -97,6 +99,104 @@ fn send_sourced_subscription_to_net_children(
     }
 }
 
+#[inline]
+fn maybe_register_local_subscriber(
+    tables: &Tables,
+    dst_face: &mut Arc<FaceState>,
+    res: &Arc<Resource>,
+    initial_interest: Option<InterestId>,
+    send_declare: &mut SendDeclare,
+) {
+    if face_hat!(dst_face).local_subs.contains_simple_resource(res) {
+        return;
+    }
+    let (should_notify, simple_interests) = match initial_interest {
+        Some(interest) => (true, HashSet::from([interest])),
+        None => face_hat!(dst_face)
+            .remote_interests
+            .iter()
+            .filter(|(_, i)| i.options.subscribers() && i.matches(res))
+            .fold(
+                (false, HashSet::new()),
+                |(_, mut simple_interests), (id, i)| {
+                    if !i.options.aggregate() {
+                        simple_interests.insert(*id);
+                    }
+                    (true, simple_interests)
+                },
+            ),
+    };
+
+    if !should_notify {
+        return;
+    }
+    let face_hat_mut = face_hat_mut!(dst_face);
+    let (_, subs_to_notify) = face_hat_mut.local_subs.insert_simple_resource(
+        res.clone(),
+        SubscriberInfo,
+        || face_hat_mut.next_id.fetch_add(1, Ordering::SeqCst),
+        simple_interests,
+    );
+
+    for update in subs_to_notify {
+        let key_expr = Resource::decl_key(
+            &update.resource,
+            dst_face,
+            push_declaration_profile(tables, dst_face),
+        );
+        send_declare(
+            &dst_face.primitives,
+            RoutingContext::with_expr(
+                Declare {
+                    interest_id: None,
+                    ext_qos: ext::QoSType::DECLARE,
+                    ext_tstamp: None,
+                    ext_nodeid: ext::NodeIdType::DEFAULT,
+                    body: DeclareBody::DeclareSubscriber(DeclareSubscriber {
+                        id: update.id,
+                        wire_expr: key_expr.clone(),
+                    }),
+                },
+                update.resource.expr().to_string(),
+            ),
+        );
+    }
+}
+
+#[inline]
+fn maybe_unregister_local_subscriber(
+    face: &mut Arc<FaceState>,
+    res: &Arc<Resource>,
+    send_ext_wire_expr: bool,
+    send_declare: &mut SendDeclare,
+) {
+    for update in face_hat_mut!(face).local_subs.remove_simple_resource(res) {
+        let ext_wire_expr = if send_ext_wire_expr {
+            WireExprType {
+                wire_expr: Resource::get_best_key(&update.resource, "", face.id),
+            }
+        } else {
+            WireExprType::null()
+        };
+        send_declare(
+            &face.primitives,
+            RoutingContext::with_expr(
+                Declare {
+                    interest_id: None,
+                    ext_qos: ext::QoSType::DECLARE,
+                    ext_tstamp: None,
+                    ext_nodeid: ext::NodeIdType::DEFAULT,
+                    body: DeclareBody::UndeclareSubscriber(UndeclareSubscriber {
+                        id: update.id,
+                        ext_wire_expr,
+                    }),
+                },
+                update.resource.expr().to_string(),
+            ),
+        );
+    }
+}
+
 #[inline]
 fn propagate_simple_subscription_to(
     tables: &mut Tables,
@@ -108,7 +208,7 @@ fn propagate_simple_subscription_to(
     send_declare: &mut SendDeclare,
 ) {
     if src_face.id != dst_face.id
-        && !face_hat!(dst_face).local_subs.contains_key(res)
+        && !face_hat!(dst_face).local_subs.contains_simple_resource(res)
         && if full_peer_net {
             dst_face.whatami == WhatAmI::Client
         } else {
@@ -118,47 +218,7 @@ fn propagate_simple_subscription_to(
                     || hat!(tables).failover_brokering(src_face.zid, dst_face.zid))
         }
     {
-        let matching_interests = face_hat!(dst_face)
-            .remote_interests
-            .values()
-            .filter(|i| i.options.subscribers() && i.matches(res))
-            .cloned()
-            .collect::<Vec<_>>();
-
-        for RemoteInterest {
-            res: int_res,
-            options,
-            ..
-        } in matching_interests
-        {
-            let res = if options.aggregate() {
-                int_res.as_ref().unwrap_or(res)
-            } else {
-                res
-            };
-            if !face_hat!(dst_face).local_subs.contains_key(res) {
-                let id = face_hat!(dst_face).next_id.fetch_add(1, Ordering::SeqCst);
-                face_hat_mut!(dst_face).local_subs.insert(res.clone(), id);
-                let key_expr =
-                    Resource::decl_key(res, dst_face, push_declaration_profile(tables, dst_face));
-                send_declare(
-                    &dst_face.primitives,
-                    RoutingContext::with_expr(
-                        Declare {
-                            interest_id: None,
-                            ext_qos: ext::QoSType::DECLARE,
-                            ext_tstamp: None,
-                            ext_nodeid: ext::NodeIdType::DEFAULT,
-                            body: DeclareBody::DeclareSubscriber(DeclareSubscriber {
-                                id,
-                                wire_expr: key_expr,
-                            }),
-                        },
-                        res.expr().to_string(),
-                    ),
-                );
-            }
-        }
+        maybe_register_local_subscriber(tables, dst_face, res, None, send_declare);
     }
 }
 
@@ -370,13 +430,6 @@ fn simple_subs(res: &Arc<Resource>) -> Vec<Arc<FaceState>> {
         .collect()
 }
 
-#[inline]
-fn remote_simple_subs(res: &Arc<Resource>, face: &Arc<FaceState>) -> bool {
-    res.session_ctxs
-        .values()
-        .any(|ctx| ctx.face.id != face.id && ctx.subs.is_some())
-}
-
 #[inline]
 fn send_forget_sourced_subscription_to_net_children(
     tables: &Tables,
@@ -426,58 +479,7 @@ fn propagate_forget_simple_subscription(
     send_declare: &mut SendDeclare,
 ) {
     for mut face in tables.faces.values().cloned() {
-        if let Some(id) = face_hat_mut!(&mut face).local_subs.remove(res) {
-            send_declare(
-                &face.primitives,
-                RoutingContext::with_expr(
-                    Declare {
-                        interest_id: None,
-                        ext_qos: ext::QoSType::DECLARE,
-                        ext_tstamp: None,
-                        ext_nodeid: ext::NodeIdType::DEFAULT,
-                        body: DeclareBody::UndeclareSubscriber(UndeclareSubscriber {
-                            id,
-                            ext_wire_expr: WireExprType::null(),
-                        }),
-                    },
-                    res.expr().to_string(),
-                ),
-            );
-        }
-        for res in face_hat!(&mut face)
-            .local_subs
-            .keys()
-            .cloned()
-            .collect::<Vec<Arc<Resource>>>()
-        {
-            if !res.context().matches.iter().any(|m| {
-                m.upgrade().is_some_and(|m| {
-                    m.context.is_some()
-                        && (remote_simple_subs(&m, &face)
-                            || remote_linkstatepeer_subs(tables, &m)
-                            || remote_router_subs(tables, &m))
-                })
-            }) {
-                if let Some(id) = face_hat_mut!(&mut face).local_subs.remove(&res) {
-                    send_declare(
-                        &face.primitives,
-                        RoutingContext::with_expr(
-                            Declare {
-                                interest_id: None,
-                                ext_qos: ext::QoSType::DECLARE,
-                                ext_tstamp: None,
-                                ext_nodeid: ext::NodeIdType::DEFAULT,
-                                body: DeclareBody::UndeclareSubscriber(UndeclareSubscriber {
-                                    id,
-                                    ext_wire_expr: WireExprType::null(),
-                                }),
-                            },
-                            res.expr().to_string(),
-                        ),
-                    );
-                }
-            }
-        }
+        maybe_unregister_local_subscriber(&mut face, res, false, send_declare);
     }
 }
 
@@ -497,7 +499,7 @@ fn propagate_forget_simple_subscription_to_peers(
             .collect::<Vec<Arc<FaceState>>>()
         {
             if face.whatami == WhatAmI::Peer
-                && face_hat!(face).local_subs.contains_key(res)
+                && face_hat!(face).local_subs.contains_simple_resource(res)
                 && !res.session_ctxs.values().any(|s| {
                     face.zid != s.face.zid
                         && s.subs.is_some()
@@ -506,24 +508,7 @@ fn propagate_forget_simple_subscription_to_peers(
                                 && hat!(tables).failover_brokering(s.face.zid, face.zid)))
                 })
             {
-                if let Some(id) = face_hat_mut!(&mut face).local_subs.remove(res) {
-                    send_declare(
-                        &face.primitives,
-                        RoutingContext::with_expr(
-                            Declare {
-                                interest_id: None,
-                                ext_qos: ext::QoSType::DECLARE,
-                                ext_tstamp: None,
-                                ext_nodeid: ext::NodeIdType::DEFAULT,
-                                body: DeclareBody::UndeclareSubscriber(UndeclareSubscriber {
-                                    id,
-                                    ext_wire_expr: WireExprType::null(),
-                                }),
-                            },
-                            res.expr().to_string(),
-                        ),
-                    );
-                }
+                maybe_unregister_local_subscriber(&mut face, res, false, send_declare);
             }
         }
     }
@@ -671,59 +656,7 @@ pub(super) fn undeclare_simple_subscription(
         }
 
         if simple_subs.len() == 1 && !router_subs && !linkstatepeer_subs {
-            let mut face = &mut simple_subs[0];
-            if let Some(id) = face_hat_mut!(face).local_subs.remove(res) {
-                send_declare(
-                    &face.primitives,
-                    RoutingContext::with_expr(
-                        Declare {
-                            interest_id: None,
-                            ext_qos: ext::QoSType::DECLARE,
-                            ext_tstamp: None,
-                            ext_nodeid: ext::NodeIdType::DEFAULT,
-                            body: DeclareBody::UndeclareSubscriber(UndeclareSubscriber {
-                                id,
-                                ext_wire_expr: WireExprType::null(),
-                            }),
-                        },
-                        res.expr().to_string(),
-                    ),
-                );
-            }
-            for res in face_hat!(face)
-                .local_subs
-                .keys()
-                .cloned()
-                .collect::<Vec<Arc<Resource>>>()
-            {
-                if !res.context().matches.iter().any(|m| {
-                    m.upgrade().is_some_and(|m| {
-                        m.context.is_some()
-                            && (remote_simple_subs(&m, face)
-                                || remote_linkstatepeer_subs(tables, &m)
-                                || remote_router_subs(tables, &m))
-                    })
-                }) {
-                    if let Some(id) = face_hat_mut!(&mut face).local_subs.remove(&res) {
-                        send_declare(
-                            &face.primitives,
-                            RoutingContext::with_expr(
-                                Declare {
-                                    interest_id: None,
-                                    ext_qos: ext::QoSType::DECLARE,
-                                    ext_tstamp: None,
-                                    ext_nodeid: ext::NodeIdType::DEFAULT,
-                                    body: DeclareBody::UndeclareSubscriber(UndeclareSubscriber {
-                                        id,
-                                        ext_wire_expr: WireExprType::null(),
-                                    }),
-                                },
-                                res.expr().to_string(),
-                            ),
-                        );
-                    }
-                }
-            }
+            maybe_unregister_local_subscriber(&mut simple_subs[0], res, false, send_declare);
         }
     }
 }
@@ -851,7 +784,7 @@ pub(super) fn pubsub_linkstate_change(
         if hat!(tables).router_peers_failover_brokering && src_face.whatami == WhatAmI::Peer {
             let to_forget = face_hat!(src_face)
                 .local_subs
-                .keys()
+                .simple_resources()
                 .filter(|res| {
                     let client_subs = res
                         .session_ctxs
@@ -868,25 +801,7 @@ pub(super) fn pubsub_linkstate_change(
                 .cloned()
                 .collect::<Vec<Arc<Resource>>>();
             for res in to_forget {
-                if let Some(id) = face_hat_mut!(&mut src_face).local_subs.remove(&res) {
-                    let wire_expr = Resource::get_best_key(&res, "", src_face.id);
-                    send_declare(
-                        &src_face.primitives,
-                        RoutingContext::with_expr(
-                            Declare {
-                                interest_id: None,
-                                ext_qos: ext::QoSType::DECLARE,
-                                ext_tstamp: None,
-                                ext_nodeid: ext::NodeIdType::default(),
-                                body: DeclareBody::UndeclareSubscriber(UndeclareSubscriber {
-                                    id,
-                                    ext_wire_expr: WireExprType { wire_expr },
-                                }),
-                            },
-                            res.expr().to_string(),
-                        ),
-                    );
-                }
+                maybe_unregister_local_subscriber(&mut src_face, &res, true, send_declare);
             }
 
             for mut dst_face in tables.faces.values().cloned() {
@@ -894,30 +809,13 @@ pub(super) fn pubsub_linkstate_change(
                     && HatTables::failover_brokering_to(links, &dst_face.zid)
                 {
                     for res in face_hat!(src_face).remote_subs.values() {
-                        if !face_hat!(dst_face).local_subs.contains_key(res) {
-                            let id = face_hat!(dst_face).next_id.fetch_add(1, Ordering::SeqCst);
-                            face_hat_mut!(&mut dst_face)
-                                .local_subs
-                                .insert(res.clone(), id);
-                            let push_declaration = push_declaration_profile(tables, &dst_face);
-                            let key_expr = Resource::decl_key(res, &mut dst_face, push_declaration);
-                            send_declare(
-                                &dst_face.primitives,
-                                RoutingContext::with_expr(
-                                    Declare {
-                                        interest_id: None,
-                                        ext_qos: ext::QoSType::DECLARE,
-                                        ext_tstamp: None,
-                                        ext_nodeid: ext::NodeIdType::default(),
-                                        body: DeclareBody::DeclareSubscriber(DeclareSubscriber {
-                                            id,
-                                            wire_expr: key_expr,
-                                        }),
-                                    },
-                                    res.expr().to_string(),
-                                ),
-                            );
-                        }
+                        maybe_register_local_subscriber(
+                            tables,
+                            &mut dst_face,
+                            res,
+                            Some(INITIAL_INTEREST_ID),
+                            send_declare,
+                        );
                     }
                 }
             }
@@ -925,139 +823,126 @@ pub(super) fn pubsub_linkstate_change(
     }
 }
 
-#[inline]
-fn make_sub_id(res: &Arc<Resource>, face: &mut Arc<FaceState>, mode: InterestMode) -> u32 {
-    if mode.future() {
-        if let Some(id) = face_hat!(face).local_subs.get(res) {
-            *id
-        } else {
-            let id = face_hat!(face).next_id.fetch_add(1, Ordering::SeqCst);
-            face_hat_mut!(face).local_subs.insert(res.clone(), id);
-            id
-        }
-    } else {
-        0
-    }
+fn get_subscribers_matching_resource<'a>(
+    tables: &'a Tables,
+    face: &Arc<FaceState>,
+    res: Option<&'a Arc<Resource>>,
+) -> impl Iterator<Item = &'a Arc<Resource>> {
+    let face_id = face.id;
+    let face_zid = face.zid;
+    let face_what_am_i = face.whatami;
+
+    hat!(tables).router_subs.iter().filter(move |sub| {
+        sub.context.is_some()
+            && res.as_ref().map(|r| sub.matches(r)).unwrap_or(true)
+            && (res_hat!(sub).router_subs.iter().any(|r| *r != tables.zid)
+                || res_hat!(sub)
+                    .linkstatepeer_subs
+                    .iter()
+                    .any(|r| *r != tables.zid)
+                || sub.session_ctxs.values().any(|s| {
+                    s.face.id != face_id
+                        && s.subs.is_some()
+                        && (s.face.whatami == WhatAmI::Client
+                            || face_what_am_i == WhatAmI::Client
+                            || (s.face.whatami == WhatAmI::Peer
+                                && hat!(tables).failover_brokering(s.face.zid, face_zid)))
+                }))
+    })
 }
 
 pub(crate) fn declare_sub_interest(
     tables: &mut Tables,
     face: &mut Arc<FaceState>,
-    id: InterestId,
+    interest_id: InterestId,
     res: Option<&mut Arc<Resource>>,
     mode: InterestMode,
     aggregate: bool,
     send_declare: &mut SendDeclare,
 ) {
-    if mode.current() {
-        let interest_id = Some(id);
-        if let Some(res) = res.as_ref() {
-            if aggregate {
-                if hat!(tables).router_subs.iter().any(|sub| {
-                    sub.context.is_some()
-                        && sub.matches(res)
-                        && (remote_simple_subs(sub, face)
-                            || remote_linkstatepeer_subs(tables, sub)
-                            || remote_router_subs(tables, sub))
-                }) {
-                    let id = make_sub_id(res, face, mode);
-                    let wire_expr =
-                        Resource::decl_key(res, face, push_declaration_profile(tables, face));
-                    send_declare(
-                        &face.primitives,
-                        RoutingContext::with_expr(
-                            Declare {
-                                interest_id,
-                                ext_qos: ext::QoSType::DECLARE,
-                                ext_tstamp: None,
-                                ext_nodeid: ext::NodeIdType::DEFAULT,
-                                body: DeclareBody::DeclareSubscriber(DeclareSubscriber {
-                                    id,
-                                    wire_expr,
-                                }),
-                            },
-                            res.expr().to_string(),
-                        ),
+    let res = res.map(|r| r.clone());
+    let mut matching_subs = get_subscribers_matching_resource(tables, face, res.as_ref());
+
+    if aggregate && (mode.current() || mode.future()) {
+        if let Some(aggregated_res) = &res {
+            let (resource_id, sub_info) = if mode.future() {
+                let face_hat_mut = face_hat_mut!(face);
+                for sub in matching_subs {
+                    face_hat_mut.local_subs.insert_simple_resource(
+                        sub.clone(),
+                        SubscriberInfo,
+                        || face_hat_mut.next_id.fetch_add(1, Ordering::SeqCst),
+                        HashSet::new(),
                     );
                 }
+                let face_hat_mut = face_hat_mut!(face);
+                face_hat_mut.local_subs.insert_aggregated_resource(
+                    aggregated_res.clone(),
+                    || face_hat_mut.next_id.fetch_add(1, Ordering::SeqCst),
+                    HashSet::from_iter([interest_id]),
+                )
             } else {
-                for sub in &hat!(tables).router_subs {
-                    if sub.context.is_some()
-                        && sub.matches(res)
-                        && (res_hat!(sub).router_subs.iter().any(|r| *r != tables.zid)
-                            || res_hat!(sub)
-                                .linkstatepeer_subs
-                                .iter()
-                                .any(|r| *r != tables.zid)
-                            || sub.session_ctxs.values().any(|s| {
-                                s.face.id != face.id
-                                    && s.subs.is_some()
-                                    && (s.face.whatami == WhatAmI::Client
-                                        || face.whatami == WhatAmI::Client
-                                        || (s.face.whatami == WhatAmI::Peer
-                                            && hat!(tables)
-                                                .failover_brokering(s.face.zid, face.zid)))
-                            }))
-                    {
-                        let id = make_sub_id(sub, face, mode);
-                        let wire_expr =
-                            Resource::decl_key(sub, face, push_declaration_profile(tables, face));
-                        send_declare(
-                            &face.primitives,
-                            RoutingContext::with_expr(
-                                Declare {
-                                    interest_id,
-                                    ext_qos: ext::QoSType::DECLARE,
-                                    ext_tstamp: None,
-                                    ext_nodeid: ext::NodeIdType::DEFAULT,
-                                    body: DeclareBody::DeclareSubscriber(DeclareSubscriber {
-                                        id,
-                                        wire_expr,
-                                    }),
-                                },
-                                sub.expr().to_string(),
-                            ),
-                        );
-                    }
-                }
-            }
-        } else {
-            for sub in &hat!(tables).router_subs {
-                if sub.context.is_some()
-                    && (res_hat!(sub).router_subs.iter().any(|r| *r != tables.zid)
-                        || res_hat!(sub)
-                            .linkstatepeer_subs
-                            .iter()
-                            .any(|r| *r != tables.zid)
-                        || sub.session_ctxs.values().any(|s| {
-                            s.subs.is_some()
-                                && (s.face.whatami != WhatAmI::Peer
-                                    || face.whatami != WhatAmI::Peer
-                                    || hat!(tables).failover_brokering(s.face.zid, face.zid))
-                        }))
-                {
-                    let id = make_sub_id(sub, face, mode);
-                    let wire_expr =
-                        Resource::decl_key(sub, face, push_declaration_profile(tables, face));
-                    send_declare(
-                        &face.primitives,
-                        RoutingContext::with_expr(
-                            Declare {
-                                interest_id,
-                                ext_qos: ext::QoSType::DECLARE,
-                                ext_tstamp: None,
-                                ext_nodeid: ext::NodeIdType::DEFAULT,
-                                body: DeclareBody::DeclareSubscriber(DeclareSubscriber {
-                                    id,
-                                    wire_expr,
-                                }),
-                            },
-                            sub.expr().to_string(),
-                        ),
-                    );
-                }
+                (0, matching_subs.next().map(|_| SubscriberInfo))
+            };
+            if mode.current() && sub_info.is_some() {
+                // send declare only if there is at least one resource matching the aggregate
+                let wire_expr = Resource::decl_key(
+                    aggregated_res,
+                    face,
+                    push_declaration_profile(tables, face),
+                );
+                send_declare(
+                    &face.primitives,
+                    RoutingContext::with_expr(
+                        Declare {
+                            interest_id: Some(interest_id),
+                            ext_qos: ext::QoSType::DECLARE,
+                            ext_tstamp: None,
+                            ext_nodeid: ext::NodeIdType::DEFAULT,
+                            body: DeclareBody::DeclareSubscriber(DeclareSubscriber {
+                                id: resource_id,
+                                wire_expr,
+                            }),
+                        },
+                        aggregated_res.expr().to_string(),
+                    ),
+                );
             }
         }
+    } else if !aggregate && mode.current() {
+        for sub in matching_subs {
+            let resource_id = if mode.future() {
+                let face_hat_mut = face_hat_mut!(face);
+                face_hat_mut
+                    .local_subs
+                    .insert_simple_resource(
+                        sub.clone(),
+                        SubscriberInfo,
+                        || face_hat_mut.next_id.fetch_add(1, Ordering::SeqCst),
+                        HashSet::from([interest_id]),
+                    )
+                    .0
+            } else {
+                0
+            };
+            let wire_expr = Resource::decl_key(sub, face, push_declaration_profile(tables, face));
+            send_declare(
+                &face.primitives,
+                RoutingContext::with_expr(
+                    Declare {
+                        interest_id: Some(interest_id),
+                        ext_qos: ext::QoSType::DECLARE,
+                        ext_tstamp: None,
+                        ext_nodeid: ext::NodeIdType::DEFAULT,
+                        body: DeclareBody::DeclareSubscriber(DeclareSubscriber {
+                            id: resource_id,
+                            wire_expr,
+                        }),
+                    },
+                    sub.expr().to_string(),
+                ),
+            );
+        }
     }
 }
 
@@ -1147,7 +1032,7 @@ impl HatPubSubTrait for HatCode {
     }
 
     fn get_subscriptions(&self, tables: &Tables) -> Vec<(Arc<Resource>, Sources)> {
-        // Compute the list of known suscriptions (keys)
+        // Compute the list of known subscriptions (keys)
         hat!(tables)
             .router_subs
             .iter()
@@ -1215,7 +1100,7 @@ impl HatPubSubTrait for HatCode {
     fn compute_data_route(
         &self,
         tables: &Tables,
-        expr: &mut RoutingExpr,
+        expr: &RoutingExpr,
         source: NodeId,
         source_type: WhatAmI,
     ) -> Arc<Route> {
@@ -1238,12 +1123,8 @@ impl HatPubSubTrait for HatCode {
                                 if net.graph.contains_node(direction) {
                                     if let Some(face) = tables.get_face(&net.graph[direction].zid) {
                                         route.insert(face.id, || {
-                                            let key_expr = Resource::get_best_key(
-                                                expr.prefix,
-                                                expr.suffix,
-                                                face.id,
-                                            );
-                                            (face.clone(), key_expr.to_owned(), source)
+                                            let wire_expr = expr.get_best_key(face.id);
+                                            (face.clone(), wire_expr.to_owned(), source)
                                         });
                                     }
                                 }
@@ -1257,32 +1138,24 @@ impl HatPubSubTrait for HatCode {
         }
 
         let mut route = RouteBuilder::new();
-        let key_expr = expr.full_expr();
-        if key_expr.ends_with('/') {
+        let Some(key_expr) = expr.key_expr() else {
             return Arc::new(route.build());
-        }
+        };
         tracing::trace!(
             "compute_data_route({}, {:?}, {:?})",
             key_expr,
             source,
             source_type
         );
-        let key_expr = match OwnedKeyExpr::try_from(key_expr) {
-            Ok(ke) => ke,
-            Err(e) => {
-                tracing::warn!("Invalid KE reached the system: {}", e);
-                return Arc::new(route.build());
-            }
-        };
-        let res = Resource::get_resource(expr.prefix, expr.suffix);
-        let matches = res
+        let matches = expr
+            .resource()
             .as_ref()
             .and_then(|res| res.context.as_ref())
             .map(|ctx| Cow::from(&ctx.matches))
-            .unwrap_or_else(|| Cow::from(Resource::get_matches(tables, &key_expr)));
+            .unwrap_or_else(|| Cow::from(Resource::get_matches(tables, key_expr)));
 
         let master = !hat!(tables).full_net(WhatAmI::Peer)
-            || *hat!(tables).elect_router(&tables.zid, &key_expr, hat!(tables).shared_nodes.iter())
+            || *hat!(tables).elect_router(&tables.zid, key_expr, hat!(tables).shared_nodes.iter())
                 == tables.zid;
 
         for mres in matches.iter() {
@@ -1324,8 +1197,12 @@ impl HatPubSubTrait for HatCode {
                 for (sid, context) in &mres.session_ctxs {
                     if context.subs.is_some() && context.face.whatami != WhatAmI::Router {
                         route.insert(*sid, || {
-                            let key_expr = Resource::get_best_key(expr.prefix, expr.suffix, *sid);
-                            (context.face.clone(), key_expr.to_owned(), NodeId::default())
+                            let wire_expr = expr.get_best_key(*sid);
+                            (
+                                context.face.clone(),
+                                wire_expr.to_owned(),
+                                NodeId::default(),
+                            )
                         });
                     }
                 }
@@ -1335,7 +1212,7 @@ impl HatPubSubTrait for HatCode {
             route.insert(mcast_group.id, || {
                 (
                     mcast_group.clone(),
-                    expr.full_expr().to_string().into(),
+                    key_expr.to_string().into(),
                     NodeId::default(),
                 )
             });
@@ -1376,9 +1253,7 @@ impl HatPubSubTrait for HatCode {
         }
 
         let mut matching_subscriptions = HashMap::new();
-        if key_expr.ends_with('/') {
-            return matching_subscriptions;
-        }
+
         tracing::trace!("get_matching_subscriptions({})", key_expr,);
 
         let res = Resource::get_resource(&tables.root_res, key_expr);
