@@ -16,5 +16,3 @@ pub(crate) mod defragmentation;
 pub(crate) mod pipeline;
 pub(crate) mod priority;
 pub(crate) mod seq_num;
-#[cfg(feature = "stats")]
-pub mod stats;
